/************************************************************************
 *   IRC - Internet Relay Chat, src/elite.c
 *   Copyright (C) 1999 Chris Wolkowski (Potvin @ dal.net)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef lint
static char sccsid[] = "@(#)elite.c     8.00 9/17/98 EliteIRCD";
#endif

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
//#include <utmp.h>
#include <stdio.h>
#include "h.h"
#include "elite.h"

time_t TSoffset = 0;

/* Hidden host code below */

#define MAXVIRTSIZE     (3 + 5 + 1)
#define HASHVAL_TOTAL   30011
#define HASHVAL_PARTIAL 211

extern  aClient me;
extern  int     seed;
int     match(char *, char *), find_exception(char *);

extern unsigned char tolowertab[];

/* I know this should be done by a standard conf function but I need
   the address before they are called */

unsigned long int
get_my_address ()
{
	/* extract the IP from the M:line in ircd.conf */
	FILE * fconf;
	char buf[512];
	char *p, *p2;
	char ipa[24];

	if ((fconf = fopen(CPATH, "r")) == NULL) {
		return 0L;
	}
	*ipa = '\0';
	while (fgets(buf, 512, fconf) != NULL) {
		if (*buf == 'M') {
			p = strchr(buf+2, ':');
			p++;
			p2 = strchr(p, ':');
			*p2 = '\0';
			strncpy (ipa, p, sizeof(ipa));
			break;
		}	
	}
	fclose (fconf);
	if (*ipa == '*')
		return 0L;
	if (*ipa)
		return inet_addr(ipa);
	return 0L;
}
				
	
int str2array(char **pparv, char *string, char *delim)
{
    char *tok;
    int pparc=0;

    tok=strtok(string,delim);
    while(tok != NULL)
    {
        pparv[pparc++]=tok;
        tok=strtok(NULL,delim);
    }

    return pparc;
}


void truncstring(char *stringvar, int firstlast, int amount){
   if (firstlast)
   {
    stringvar+=amount;
    *stringvar=0;
    stringvar-=amount;
   }
    else
   {
    stringvar+=strlen(stringvar);
    stringvar-=amount;
   }
}

int Maskchecksum (char *data, int len)
{
        int                     i;
        int                     j;

        j=0;
        for (i=0 ; i<len ; i++)
        {
          j += *data++ * (i < 16 ? (i+1)*(i+1) : i*(i-15));

        }

        return (j+1000)%0xffff;
}

unsigned z(char *s)
{
    unsigned int i;
        i = Maskchecksum (s, strlen(s));
    return i;
}

/* Regular user host */
void    make_virthost(char *curr, char *new)
{
    unsigned hash_total = z(curr) % HASHVAL_TOTAL, hash_partial;
    char     destroy[HOSTLEN+1], *parv[HOSTLEN+1], tmpnew[HOSTLEN], *ptr;
    int      parc=0, len=0, overflow=0, exception=0;

    strncpyzt(destroy, curr, HOSTLEN);

    len = strlen(destroy);

    if ((len + MAXVIRTSIZE) > HOSTLEN)
    {
        overflow = (len + MAXVIRTSIZE) - HOSTLEN;

        ptr = &destroy[overflow];
    }
    else
        ptr = &destroy[0];

    parc = str2array(parv, ptr, ".");

    if (parc == 4)
    {
        len = strlen(parv[3]);

        if (strchr("0123456789", parv[3][len-1]))
        {
            hash_partial = (z(parv[0])+z(parv[1])+z(parv[2])) % HASHVAL_PARTIAL;
            sprintf(tmpnew, "%s.%s.%u.%s-%u",
                parv[0], parv[1], hash_partial, hidden_host, hash_total);

            strncpyzt(new, tmpnew, HOSTLEN);
            return;
        }
    }

    if (parc == 2)
    {
        sprintf(tmpnew, "%s-%u.%s.%s",
            hidden_host, hash_total, parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }

    len = strlen(parv[parc-1]);

    if ((len==2) && (find_exception(parv[parc-1])))
        exception = 1;

    if ((len == 2) && (exception == 0))
    {
        sprintf(tmpnew, "%s-%u.%s.%s.%s",
            hidden_host, hash_total, parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }
    sprintf(tmpnew, "%s-%u.%s.%s", hidden_host, hash_total, parv[parc-2], parv[parc-1]);
    strncpyzt(new, tmpnew, HOSTLEN);
    return;
}

/* Virtual host */
void    make_vhost(char *new, char *vhostname)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", vhostname);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Netadmin host */
void	make_netadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", netadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Coadmin host */
void    make_coadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", coadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Techadmin host */
void    make_techadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", techadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Server admin host */
void    make_adminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", admin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Server coadmin host */
void    make_scoadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", scoadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Service admin host */
void    make_sadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", sadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Global Oper host */
void    make_operhost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", oper_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}
/* Local Oper host */
void    make_locophost(char *new)
{
    char     tmpnew[HOSTLEN];

        sprintf(tmpnew, "%s", locop_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/*
** m_newhost
*/

int     m_newhost(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[];
    {
        Reg3    anUser  *user;
        char    *newhost;
        int     what, setflags;
        char 	*s;
        int     legalhost = 1;

         if (check_registered(sptr))
                 return 0;

	if (NEWHOST==FALSE) {
	     sendto_one(sptr, ":%s NOTICE %s :*** /NEWHOST is OFF.", me.name, parv[0]);
                return 0;
	} else {

        if (!IsOper(sptr)) {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0; }

        newhost = parc > 1 ? parv[1] : NULL;

        if (BadPtr(newhost))
            {
     sendto_one(sptr, ":%s NOTICE %s :*** Syntax: /NEWHOST <new host>", me.name, parv[0]);
                return 0;
            }

            if (strlen(parv[1]) > (HOSTLEN)) {
           sendto_one(sptr, ":%s NOTICE %s :*** Error: Hostname too long", me.name, parv[0]);
           return;
            }

                for (s = parv[1]; *s; s++) {
                        if (!isallowed(*s)) {
                                legalhost = 0;
                        }
                }
                if (legalhost == 0) {
                        sendto_one(sptr, ":%s NOTICE %s :*** Error: The hostname containes \
			illegal characters, please only use a-z, A-Z, 0-9, '-' & '.'", me.name, 
			parv[0]);
                        return;
                }

                            {
        sendto_umode(UMODE_EYES, "*** Hostname: \2%s (%s@%s)\2  -->  \2%s (%s@%s)\2", parv[0],
                        sptr->user->username, IsHidden(sptr) ? sptr->user->virthost : 
			sptr->user->host, parv[0], sptr->user->username, parv[1]);
        sptr->umodes |= UMODE_HIDE;
	(void)sprintf(sptr->user->virthost, "%s", newhost);
	sendto_serv_butone(cptr, ":%s NEWHOST %s", parv[0], newhost);
	SetHidden(sptr);
	if (MyClient(sptr)) {
	sendto_one(sptr, ":%s NOTICE %s :*** New hostmask: \2%s (%s@%s)\2", me.name, parv[0],
							parv[0], sptr->user->username, parv[1]);
	sendto_one(sptr, ":%s NOTICE %s :*** To disable  : \2[/mode %s -x]\2", me.name, parv[0], 
							parv[0]);
	return 0;
			}
          }
    }
}

/*
** m_display
**      parv[0] = sender prefix
**      parv[1] = filename
**	parv[2] = servername (netadmins only)
*/
int     m_display(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[];
{
        int     fd, nr;
        char    line[80];
        Reg1    char     *tmp;
        struct  stat    sb;
        struct  tm      *tm;

        if (check_registered(sptr))
                return 0;

        if (!IsAdmin(sptr)) {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0; }

        if (parc < 2)
        {
                sendto_one (sptr, err_str(ERR_NEEDMOREPARAMS),
                            me.name, parv[0], "DISPLAY");
                return 0;
        }

	if (IsNetAdmin(sptr)) {
        if (hunt_server(cptr, sptr, ":%s DISPLAY :%s", 2,parc,parv)!=HUNTED_ISME)
                return 0;
        }

        /*
         * stop NFS hangs...most systems should be able to open a file in
         * 3 seconds. -avalon (curtesy of wumpus)
         */

	if (strstr(parv[1], "/")) { 
	sendto_one(sptr, ":%s NOTICE %s :*** Sorry, only files in the current directory.", me.name, parv[0]);
		return 0; }

        fd = open(parv[1], O_RDONLY);
        if (fd == -1)
            {
		sendto_one(sptr, ":%s NOTICE %s :*** Sorry, could not open [%s] - Not enough access or the file does not exist.", me.name, parv[0], parv[1]);
                return 0;
            }
        (void)fstat(fd, &sb);
        sendto_umode(UMODE_EYES,"*** %s (%s@%s) displayed file: \2%s\2", parv[0],
        sptr->user->username, IsHidden(sptr) ? sptr->user->virthost : sptr->user->host, parv[1]);
	sendto_one(sptr, ":%s NOTICE %s :-- DISPLAYING [\2%s\2] --", me.name, parv[0], parv[1]);
        (void)dgets(-1, NULL, 0); /* make sure buffer is at empty pos */
        while ((nr=dgets(fd, line, sizeof(line)-1)) > 0)
            {
                line[nr]='\0';
                if ((tmp = (char *)index(line,'\n')))
                        *tmp = '\0';
                if ((tmp = (char *)index(line,'\r')))
                        *tmp = '\0';
        sendto_one(sptr, ":%s NOTICE %s :%s", me.name, parv[0], line);
            }
        (void)dgets(-1, NULL, 0); /* make sure buffer is at empty pos */
	sendto_one(sptr, ":%s NOTICE %s :-- CLOSING [\2%s\2] -- ", me.name, parv[0], parv[1]);
        (void)close(fd);
        return 0;
    }

/*
** m_edit
**      parv[0] = sender prefix
**      parv[1] = filename
**      parv[2] = servername (netadmins only)
**      parv[3] = stuff to write
*/
int     m_edit(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[];
{
        int     i;
	FILE	*fd;

        if (check_registered(sptr))
                return 0;

        if (!IsAdmin(sptr)) {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0; }

        if (IsNetAdmin(sptr)) {

        if (hunt_server(cptr, sptr, ":%s EDIT %s %s :%s", 2,parc,parv)!=HUNTED_ISME)
                return 0;

        if (parc < 4)
        {
                sendto_one (sptr, err_str(ERR_NEEDMOREPARAMS),
                            me.name, parv[0], "EDIT");
                return 0;
        }

        /*
         * stop NFS hangs...most systems should be able to open a file in
         * 3 seconds. -avalon (curtesy of wumpus)
         */

        if (strstr(parv[1], "/")) {
        sendto_one(sptr, ":%s NOTICE %s :*** Sorry, only in the current directory.", me.name, parv[0]);
                return 0; }

        fd = fopen(parv[1], "a");
        if (fd == NULL)
            {
                sendto_one(sptr, ":%s NOTICE %s :*** Sorry, could not open [%s] - Not enough access or the file does not exist.", me.name, parv[0], parv[1]);
                return 0;
            }
#ifdef _WIN32
        for (i=3; i<parc; i++)
	{
		if (i!=parc-1) {
		fprintf (fd, "\n");
		fprintf (fd, "%s ", parv[i]);
		} else {
		fprintf (fd, "\n");
		fprintf (fd, "%s ", parv[i]); }
	}
#else
        for (i=3; i<parc; i++)
	{
		if (i!=parc-1)
        fprintf (fd, "%s ", parv[i]);
		else
		fprintf (fd, "%s\n", parv[i]);
	}
#endif

        sendto_one(sptr, ":%s NOTICE %s :*** Wrote to file: \2%s\2, server: \2%s\2", me.name, parv[0], parv[1], parv[2]);
        sendto_umode(UMODE_EYES,"*** %s (%s@%s) wrote to file: \2%s\2", parv[0],
        sptr->user->username, IsHidden(sptr) ? sptr->user->virthost : sptr->user->host, parv[1]);
	} else {

        if (parc < 3)
        {
                sendto_one (sptr, err_str(ERR_NEEDMOREPARAMS),
                            me.name, parv[0], "EDIT");
                return 0;
        }

        /*
         * stop NFS hangs...most systems should be able to open a file in
         * 3 seconds. -avalon (curtesy of wumpus)
         */

        if (strstr(parv[1], "/")) {
        sendto_one(sptr, ":%s NOTICE %s :*** Sorry, only in the current directory.", me.name, parv[0]);
                return 0; }

        fd = fopen(parv[1], "a");
        if (fd == NULL)
            {
                sendto_one(sptr, ":%s NOTICE %s :*** Sorry, could not open [%s] - Not enough access or the file does not exist.", me.name, parv[0], parv[1]);
                return 0;
            }

#ifdef _WIN32
        for (i=2; i<parc; i++)
	{
		if (i!=parc-1) {
		fprintf (fd, "\n");
		fprintf (fd, "%s ", parv[i]);
		} else {
		fprintf (fd, "\n");
		fprintf (fd, "%s ", parv[i]); }
	}
#else
        for (i=2; i<parc; i++)
	{
		if (i!=parc-1)
        fprintf (fd, "%s ", parv[i]);
		else
		fprintf (fd, "%s\n", parv[i]);
	}
#endif

        sendto_one(sptr, ":%s NOTICE %s :*** Wrote to file: \2%s\2", me.name, parv[0], parv[1]);
	sendto_umode(UMODE_EYES,"*** %s (%s@%s) wrote to file: \2%s\2", parv[0],
	sptr->user->username, IsHidden(sptr) ? sptr->user->virthost : sptr->user->host, parv[1]);
	}
	fclose (fd);
        return 0;
    }

/*
** m_announce
**      parv[0] = sender prefix
**      parv[1] = announcement
*/
int     m_announce(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[];
{
	int	i, fp, nr;
	FILE	*fd;
        char    line[80];
        Reg1    char     *tmp;
        struct  stat    sb;
        struct  tm      *tm;
	char ifilename[ 512 ] = "svrmsg";

        if (check_registered(sptr))
                return 0;

        if (!IsOper(sptr)) {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0; }

        if (parc < 2)
        {
                sendto_one (sptr, err_str(ERR_NEEDMOREPARAMS),
                            me.name, parv[0], "ANNOUNCE");
                return 0;
        }

	fd = fopen(ifilename, "a");
	if (fd == NULL)
		return 0;
		
        for (i=1; i<parc; i++)
	{
		if (i!=parc-1)
        fprintf (fd, "%s ", parv[i]);
                else
        fprintf (fd, "%s\n", parv[i]);
        }
	fclose (fd);

	fp = open(ifilename, O_RDONLY);
	if (fp == -1)
		return 0;

	(void)fstat(fp, &sb);
	(void)dgets(-1, NULL, 0); /* make sure buffer is at empty pos */
        while ((nr=dgets(fp, line, sizeof(line)-1)) > 0)
            {
                line[nr]='\0';
                if ((tmp = (char *)index(line,'\n')))
                        *tmp = '\0';
                if ((tmp = (char *)index(line,'\r')))
			*tmp = '\0';
	sendto_all("*** [\2Server Message\2] - %s", line);
            }
        (void)dgets(-1, NULL, 0); /* make sure buffer is at empty pos */
	(void)close(fp);
	remove(ifilename);
	return 0;
}

int     m_vhost(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[];
    {
        FILE    *file;
        char    buffer[2048];
        int     len;
        char    *pass, *vhostname = NULL;
        aClient *acptr;
	aConfItem *aconf;
        anUser  *user = sptr->user;

        if (check_registered(sptr))
                return 0;

        if (parc < 2)
        {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "VHOST");
                return 0;
        }

	if (!strcasecmp(parv[1], "on")) {
        if (parc < 4)
        {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "VHOST");
                return 0;
        }

        if (!(file = fopen(VHOSTFILE, "r" )))
        {
        sendto_one(sptr, ":%s NOTICE %s :*** Error: Could not open file %s.", me.name, parv[0], VHOSTFILE);
        return EXIT_FAILURE;
        }
        while (1)
        {
        fgets(buffer, 2048, file);
        /* eof? - failed to find a match */
        if (feof(file))
        {
            sendto_one(sptr, err_str(ERR_LOGINPASSWRONG), me.name, parv[0]);
            fclose(file);
            return 0;
        }

        /* overpunch '\n' */
        len = strlen(buffer);
        buffer[len - 1] = 0;

        /* find and overpunch the ':' */
        pass = strchr(buffer, ':');
        if (pass)
            *pass++ = 0;
        else
            continue;
        vhostname = strchr(pass, ':');
        if (vhostname)
            *vhostname++ = 0;
        else
            continue;

        /* check the login and password */
        if (!strcmp(buffer, parv[2]) && !strcmp(pass, parv[3]))
        {
            sendto_one(sptr, ":%s NOTICE %s :*** VHOST [%s] : ON", me.name, parv[0],
                                        vhostname);

        sendto_umode(UMODE_EYES, "*** VHOST: \2%s (%s@%s)\2  -->  \2%s (%s@%s)\2", parv[0],
                        sptr->user->username, IsHidden(sptr) ? sptr->user->virthost :
                        sptr->user->host, parv[0], sptr->user->username, vhostname);

	        sptr->umodes |= UMODE_HIDE;
		SetHidden(sptr);
                make_vhost(user->virthost, vhostname);
		sendto_serv_butone(cptr, ":%s SENDHOST %s", parv[0], vhostname);

            fclose(file);
            return 1;
            }
        }
    fclose(file);
    return 0;

	} else if (!strcasecmp(parv[1], "off")) {

            sendto_one(sptr, ":%s NOTICE %s :*** VHOST [%s] : OFF", me.name, parv[0],
					user->virthost);

                sptr->umodes &= ~UMODE_HIDE;
                ClearHidden(sptr);
                make_vhost(user->virthost, sptr->sockhost);
                sendto_serv_butone(cptr, ":%s SENDHOST %s", parv[0], sptr->sockhost);

	} else {
	sendto_one(sptr, ":%s NOTICE %s :*** VHOST: Invalid option. [ON/OFF]", me.name, parv[0]);
	return 0;
 }
}

/*
** m_sendhost
*/

int     m_sendhost(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[];
{
        Reg3    anUser  *user;
        char    *vhost;

	if (!IsServer(cptr))
		return 0;

        vhost = parc > 1 ? parv[1] : NULL;

                sptr->umodes |= UMODE_HIDE;
                SetHidden(sptr);
        (void)sprintf(sptr->user->virthost, "%s", vhost);
        sendto_serv_butone(cptr, ":%s SENDHOST %s", parv[0], vhost);
        return 0;
}
